# -*- coding: utf-8 -*-
import getpass
import json
import netrc
import os
import time
import stat
import shutil
import tarfile
import urlparse

import git
import requests
import yaml

from .serialize import register_yaml_extensions, Trackable, File
from .git_sync import GitSyncHandler
from .sync_helpers import (
    extra_git_kwargs, git_update_gitignore, git_pull_develop_bundle)
from .utils import (
    validate_boilerplate_config, bundle_boilerplate, filter_template_files,
    filter_static_files, validate_app_config, bundle_app, filter_sass_files,
    resource_path, cli_confirm, load_app_config, load_boilerplate_config)


CACERT_PEM_PATH = resource_path('cacert.pem')


class WritableNetRC(netrc.netrc):
    def __init__(self, *args, **kwargs):
        netrc_path = self.get_netrc_path()
        if not os.path.exists(netrc_path):
            open(netrc_path, 'a').close()
            os.chmod(netrc_path, 0600)
        # XXX netrc uses os.environ['HOME'] which isn't defined on Windows
        os.environ['HOME'] = os.path.expanduser('~')
        netrc.netrc.__init__(self, *args, **kwargs)

    def get_netrc_path(self):
        home = os.path.expanduser('~')
        return os.path.join(home, '.netrc')

    def add(self, host, login, account, password):
        self.hosts[host] = (login, account, password)

    def remove(self, host):
        if host in self.hosts:
            del self.hosts[host]

    def write(self, path=None):
        if path is None:
            path = self.get_netrc_path()
        with open(path, 'w') as fobj:
            for machine, data in self.hosts.items():
                login, account, password = data
                fobj.write('machine %s\n' % machine)
                if login:
                    fobj.write('\tlogin %s\n' % login)
                if account:
                    fobj.write('\taccount %s\n' % account)
                if password:
                    fobj.write('\tpassword %s\n' % password)


class SingleHostSession(requests.Session):
    def __init__(self, host, **kwargs):
        super(SingleHostSession, self).__init__()
        self.host = host.rstrip('/')
        for key, value in kwargs.items():
            setattr(self, key, value)

    def request(self, method, url, *args, **kwargs):
        url = self.host + url
        # Use local copy of 'cacert.pem' for easier packaging
        kwargs['verify'] = CACERT_PEM_PATH
        return super(SingleHostSession, self).request(method, url, *args, **kwargs)


class Client(object):
    APP_FILENAME_JSON = 'app.json'
    APP_FILENAME_YAML = 'app.yaml'
    ADDON_FILENAME_JSON = 'addon.json'
    ADDON_FILENAME_YAML = 'addon.yaml'
    BOILERPLATE_FILENAME_JSON = 'boilerplate.json'
    BOILERPLATE_FILENAME_YAML = 'boilerplate.yaml'
    ALDRYN_CONFIG_FILENAME = 'aldryn_config.py'
    ALDRYN_DOT_FILENAME = '.aldryn'
    ALDRYN_HOST_DEFAULT = 'https://control.aldryn.com'
    ALDRYN_HOST_KEY = 'ALDRYN_HOST'
    ALDRYN_SYNC_LOCK_FILENAME = '.aldryn-sync-lock'
    DATA_FILENAME = 'data.yaml'
    SETUP_FILENAME = 'setup.py'

    # messages
    DIRECTORY_ALREADY_SYNCING_MESSAGE = 'Directory already syncing.'
    NETWORK_ERROR_MESSAGE = 'Network error.\nPlease check your connection and try again later.'
    PROTECTED_FILE_CHANGE_MESSAGE = 'You are overriding file "%s".\nThis file is protected by the boilerplate.'
    SYNC_NETWORK_ERROR_MESSAGE = "Couldn't sync changes.\nPlease check your connection and try again later."

    ALL_CONFIG_FILES = [
        APP_FILENAME_JSON, APP_FILENAME_YAML,
        BOILERPLATE_FILENAME_JSON, BOILERPLATE_FILENAME_YAML,
        ALDRYN_CONFIG_FILENAME, SETUP_FILENAME, DATA_FILENAME]

    def __init__(self, host, interactive=True):
        register_yaml_extensions()
        self.host = urlparse.urlparse(host)[1]
        self.interactive = interactive
        self.netrc = WritableNetRC()
        auth_data = self.get_auth_data()
        if auth_data:
            headers = {
                'Authorization': 'Basic %s' % auth_data[2]
            }
        else:
            headers = {}
        self.session = SingleHostSession(
            host, headers=headers, trust_env=False)

        self._sync_handlers_cache = {}
        print self.host

    def get_auth_data(self):
        return self.netrc.hosts.get(self.host)

    def is_logged_in(self):
        auth_data = self.get_auth_data()
        return bool(auth_data)

    def get_login(self):
        if self.is_logged_in():
            auth_data = self.get_auth_data()
            return auth_data[0]

    def logout(self):
        while self.interactive:
            answer = raw_input('Are you sure you want to continue? [yN]')
            if answer.lower() == 'n' or not answer:
                print "Aborted"
                return True
            elif answer.lower() == 'y':
                break
            else:
                print "Invalid answer, please type either y or n"
        self.netrc.remove(self.host)
        self.netrc.write()

    def login(self, email=None, password=None):
        if email is None:
            email = raw_input('E-Mail: ')
        if password is None:
            password = getpass.getpass('Password: ')
        try:
            response = self.session.post(
                '/api/v1/login/', data={'email': email, 'password': password})
        except (requests.exceptions.ConnectionError,
                requests.exceptions.Timeout):
            return (False, Client.NETWORK_ERROR_MESSAGE)
        if response.ok:
            token = response.content
            self.session.headers = {
                'Authorization': 'Basic %s' % token
            }
            self.netrc.add(self.host, email, None, token)
            self.netrc.write()
            msg = "Logged in as %s" % email
            return (True, msg)
        elif response.status_code == requests.codes.forbidden:
            if response.content:
                msg = response.content
            else:
                msg = "Could not log in, invalid email or password"
            return (False, msg)
        else:
            msgs = []
            if response.content and response.status_code < 500:
                msgs.append(response.content)
            msgs.append("There was a problem logging in, please try again later.")
            return (False, '\n'.join(msgs))

    def upload_boilerplate(self, path='.'):
        is_loaded, result = load_boilerplate_config(path)
        if is_loaded:
            config, extra_file_paths = result
        else:
            return (False, result)
        data_filename = os.path.join(path, Client.DATA_FILENAME)
        if os.path.exists(data_filename):
            with Trackable.tracker as extra_objects:
                with open(data_filename) as fobj2:
                    data = yaml.safe_load(fobj2)
                extra_file_paths.extend([f.path for f in extra_objects[File]])
        else:
            data = {}
        is_valid, error_msg = validate_boilerplate_config(config, path)
        if not is_valid:
            return (False, error_msg)
        tarball = bundle_boilerplate(config, data, path, extra_file_paths, templates=filter_template_files,
                                     static=filter_static_files, private=filter_sass_files)
        try:
            response = self.session.post('/api/v1/boilerplates/', files={'boilerplate': tarball})
        except (requests.exceptions.ConnectionError,
                requests.exceptions.Timeout):
            return (False, Client.NETWORK_ERROR_MESSAGE)
        msg = '\t'.join([str(response.status_code), response.content])
        return (True, msg)

    def validate_boilerplate(self, path='.'):
        is_loaded, result = load_boilerplate_config(path)
        if is_loaded:
            config, extra_file_paths = result
            return validate_boilerplate_config(config, path)
        else:
            return (False, result)

    def upload_app(self, path='.'):
        is_loaded, result = load_app_config(path)
        if is_loaded:
            config = result
        else:
            return (False, result)
        is_valid, msg = validate_app_config(config, path)
        if not is_valid:
            return (False, msg)
        setup_filename = os.path.join(path, Client.SETUP_FILENAME)
        if not os.path.exists(setup_filename):
            msg = "File '%s' not found." % Client.SETUP_FILENAME
            return (False, msg)
        aldryn_config_filename = os.path.join(
            path, Client.ALDRYN_CONFIG_FILENAME)
        msgs = []
        if os.path.exists(aldryn_config_filename):
            with open(aldryn_config_filename) as fobj:
                script = fobj.read()
        else:
            script = ''
            msgs.append("Warning: File '%s' not found." % Client.ALDRYN_CONFIG_FILENAME)
            msgs.append("Your app will not have any configurable settings.")
        tarball = bundle_app(config, script, path)
        try:
            response = self.session.post('/api/v1/apps/', files={'app': tarball})
        except (requests.exceptions.ConnectionError,
                requests.exceptions.Timeout):
            return (False, Client.NETWORK_ERROR_MESSAGE)
        msgs.append('\t'.join([str(response.status_code), response.content]))
        return (True, '\n'.join(msgs))

    def validate_app(self, path='.'):
        setup_filename = os.path.join(path, Client.SETUP_FILENAME)
        if not os.path.exists(setup_filename):
            msg = "File '%s' not found." % Client.SETUP_FILENAME
            return (False, msg)
        is_loaded, result = load_app_config(path)
        if is_loaded:
            config = result
            return validate_app_config(config, path)
        else:
            return (False, result)

    def _acquire_sync_lock(self, sync_dir):
        lock_filename = os.path.join(
            sync_dir, Client.ALDRYN_SYNC_LOCK_FILENAME)
        try:
            fd = os.open(lock_filename, os.O_CREAT | os.O_EXCL)
            os.close(fd)
        except OSError:
            return False
        else:
            return True

    def _remove_sync_lock(self, sync_dir):
        lock_filename = os.path.join(
            sync_dir, Client.ALDRYN_SYNC_LOCK_FILENAME)
        if os.path.exists(lock_filename):
            os.remove(lock_filename)

    def sync(self, network_error_callback, sync_error_callback,
             protected_file_change_callback, sitename=None, path='.', force=False,
             sync_indicator_callback=None, stop_sync_callback=None):
        aldryn_dot_filename = os.path.join(path, Client.ALDRYN_DOT_FILENAME)
        if not sitename:
            if os.path.exists(aldryn_dot_filename):
                with open(aldryn_dot_filename, 'r') as fobj:
                    sitename = fobj.read().strip()
            if not sitename:
                msg = "Please specify a sitename using --sitename."
                return (False, msg)
        if '.' in sitename:
            sitename = sitename.split('.')[0]
        git_sync_params = {}
        if not os.path.exists(path):
            os.mkdir(path)
        if os.listdir(path):  # dir isn't empty
            git_dir = os.path.join(path, '.git')
            if os.path.exists(aldryn_dot_filename) and os.path.exists(git_dir):
                with open(aldryn_dot_filename, 'r') as fobj:
                    dir_sitename = fobj.read().strip()
                if sitename != dir_sitename:
                    msg = 'This directory is already being synced with the website "%s"' % dir_sitename
                    return (False, msg)
                repo = git.Repo(path)
                last_synced_commit = repo.git.execute(
                    ['git', 'rev-parse', 'develop_bundle/develop'],
                    **extra_git_kwargs)
                git_sync_params['last_synced_commit'] = last_synced_commit
            else:
                msg = ('The folder you selected is not empty. '
                       'Please choose an empty folder to sync your files.')
                return (False, msg)
        else:
            g = git.Git(path)
            g.execute(['git', 'init'], **extra_git_kwargs)
            repo = git.Repo(path)
            cfg = repo.config_writer()
            cfg.set_value('user', 'name', 'Git Sync by')
            cfg.set_value('user', 'email', self.get_login())
            cfg.write()
            cfg._lock._release_lock()
            del cfg

        if not self._acquire_sync_lock(path):
            if self.interactive:
                if not cli_confirm(
                        'Are you sure you want to start syncing anyway?',
                        message='It seems that you are already syncing this directory.',
                        default=False):
                    return (False, 'Aborted')
            else:
                if force:
                    pass
                else:
                    return (False, Client.DIRECTORY_ALREADY_SYNCING_MESSAGE)

        try:
            response = self.session.get(
                '/api/v1/git-sync/%s/' % sitename, params=git_sync_params, stream=True,
                headers={'accept': 'application/octet'})
        except (requests.exceptions.ConnectionError,
                requests.exceptions.Timeout):
            self._remove_sync_lock(path)
            return (False, Client.NETWORK_ERROR_MESSAGE)

        def report_unexpected_http_response(response):
            msgs = []
            msgs.append("Unexpected HTTP Response %s" % response.status_code)
            if response.status_code < 500:
                msgs.append(response.content)
            self._remove_sync_lock(path)
            return (False, '\n'.join(msgs))

        upstream_modified = True
        if response.status_code == 304:
            upstream_modified = False
        elif response.status_code != 200:
            return report_unexpected_http_response(response)

        if upstream_modified:
            try:
                git_pull_develop_bundle(response, repo, path)
            except git.exc.GitCommandError as e:
                # there were changes both in the upstream and uncommited local ones
                # next auto-merge will resolve the conflict (preferring our commits)
                # and then applying the upstream changes
                print e
        git_update_gitignore(repo, ['.*', '!.gitignore', '/db_dumps/'])

        try:
            response = self.session.get(
                '/api/v1/sync/%s/unchanged-protected-files' % sitename,
                headers={'accept': 'application/json'})
        except (requests.exceptions.ConnectionError,
                requests.exceptions.Timeout):
            self._remove_sync_lock(path)
            return (False, Client.NETWORK_ERROR_MESSAGE)

        if response.status_code != 200:
            return report_unexpected_http_response(response)

        protected_files = response.json()
        # making the protected files read-only
        for rel_protected_path in protected_files:
            protected_path = os.path.join(path, rel_protected_path)
            mode = os.stat(protected_path)[stat.ST_MODE]
            read_only_mode = mode & ~stat.S_IWUSR & ~stat.S_IWGRP & ~stat.S_IWOTH
            os.chmod(protected_path, read_only_mode)

        # successfully initialized the sync, saving the site's name
        with open(aldryn_dot_filename, 'w') as fobj:
            fobj.write(sitename)

        upstream_remote = repo.remotes.develop_bundle
        last_synced_commit = upstream_remote.refs.develop.commit.hexsha

        sync_handler = GitSyncHandler(
            self, sitename, repo, last_synced_commit,
            network_error_callback, sync_error_callback,
            protected_files, protected_file_change_callback,
            relpath=path, sync_indicator_callback=sync_indicator_callback)
        sync_handler.start()
        self._sync_handlers_cache[sitename] = sync_handler
        if self.interactive:
            print "Done, now watching for changes. You can stop the sync by hitting Ctrl-c in this shell"

            try:
                while self._is_syncing(sitename):
                    time.sleep(1)
            except KeyboardInterrupt:
                sync_handler.stop()
            return (True, 'Stopped syncing')
        else:
            return (True, sync_handler)

    def _is_syncing(self, sitename):
        return sitename in self._sync_handlers_cache

    def create_workspace(self, sitename, path=None):
        if path is None:
            path = './%s' % sitename
        path = os.path.abspath(path)
        success, msg = self.workspace_download_site(sitename, path=path)
        if not success:
            return success, msg
        success, msg = self.workspace_init_virtualenv(sitename, path)
        if not success:
            return success, msg
        return True, sitename

    def workspace_download_site(self, sitename, path):
        if not os.path.exists(path):
            os.mkdir(path)
        site_path = os.path.join(path, '.site')
        try:
            response = self.session.get(
                '/api/v1/workspace/%s/download/' % sitename, stream=True,
                headers={'accept': 'application/x-tar-gz'})
        except (requests.exceptions.ConnectionError,
                requests.exceptions.Timeout):
            return (False, Client.NETWORK_ERROR_MESSAGE)
        if response.status_code != 200:
            msgs = []
            msgs.append("Unexpected HTTP Response %s" % response.status_code)
            if response.status_code < 500:
                msgs.append(response.content)
            return (False, '\n'.join(msgs))
        else:
            # unpack the downloaded tar.gz containing the whole project
            if os.path.exists(site_path):
                print "deleting old .site"
                shutil.rmtree(site_path)
                if os.path.exists(site_path):
                    return (False, "Failed to delete old .site directory")

            print "extracting site to .site"
            from StringIO import StringIO
            tar = tarfile.open(mode='r:gz', fileobj=StringIO(response.content))
            tar.extractall(path=site_path)
            tar.close()
            print "finished extracting site"
            return True, sitename

    def workspace_init_virtualenv(self, sitename, path):
        virtualenv_path = os.path.join(path, '.virtualenv')
        requirements_path = os.path.join(path, '.site/requirements.txt')
        pip_path = os.path.join(path, '.virtualenv/bin/pip')
        import subprocess
        subprocess.call(['virtualenv', virtualenv_path])

        subprocess.call([pip_path, 'install', '-r', requirements_path])
        with open(os.path.join(virtualenv_path, 'lib/python2.7/site-packages/aldrynsite.pth'), 'w+') as fobj:
            fobj.write(os.path.abspath('.site/'))
        return True, sitename

    def sites(self):
        try:
            response = self.session.get('/api/v1/sites/', stream=True)
        except (requests.exceptions.ConnectionError,
                requests.exceptions.Timeout):
            return (False, Client.NETWORK_ERROR_MESSAGE)
        if response.status_code != 200:
            msgs = []
            msgs.append("Unexpected HTTP Response %s" % response.status_code)
            if response.status_code < 500:
                msgs.append(response.content)
            return (False, '\n'.join(msgs))
        else:
            sites = json.loads(response.content)
            if self.interactive:
                data = json.dumps(sites, sort_keys=True, indent=4, separators=(',', ': '))
            else:
                data = sites
            return (True, data)

    def newest_version(self):
        try:
            response = self.session.get('/api/v1/newest-client-app-version/')
        except (requests.exceptions.ConnectionError,
                requests.exceptions.Timeout):
            return (False, Client.NETWORK_ERROR_MESSAGE)
        if response.status_code == 404:
            return (True, None)
        elif response.status_code != 200:
            msgs = []
            msgs.append("Unexpected HTTP Response %s" % response.status_code)
            if response.status_code < 500:
                msgs.append(response.content)
            return (False, '\n'.join(msgs))
        else:
            version = json.loads(response.content)
            return (True, version)



# -*- coding: utf-8 -*-
import logging
import logging.handlers
import os

###############################################################################
# Logging configuration
###############################################################################
BACKUP_COUNT = 2
FORMAT = '%(asctime)s|%(name)s|%(levelname)s: %(message)s'
LOG_FILENAME = '.sync.log'
MAX_BYTES = 100 * (2 ** 10)  # 100KB
SYNCABLE_DIRECTORIES = ('templates/', 'static/', 'private/')


extra_git_kwargs = {}
if os.name == 'nt':
        import subprocess
        startupinfo = subprocess.STARTUPINFO()
        startupinfo.dwFlags |= subprocess.STARTF_USESHOWWINDOW
        extra_git_kwargs['startupinfo'] = startupinfo
        istream = subprocess.PIPE
        extra_git_kwargs['istream'] = istream


def get_site_specific_logger(sitename, site_dir):
    log_filename = os.path.join(site_dir, LOG_FILENAME)
    rotating_handler = logging.handlers.RotatingFileHandler(
        log_filename, maxBytes=MAX_BYTES, backupCount=BACKUP_COUNT)
    formatter = logging.Formatter(FORMAT)
    rotating_handler.setFormatter(formatter)
    rotating_handler.setLevel(logging.DEBUG)

    sync_logger = logging.getLogger(sitename)
    sync_logger.setLevel(logging.DEBUG)
    sync_logger.handlers = []
    sync_logger.addHandler(rotating_handler)
    return sync_logger
###############################################################################


def git_changes(repo):
    added = []
    deleted = []
    other = []
    git_status = repo.git.execute(
        ['git', 'status', '--porcelain', '--untracked-files=all'],
        **extra_git_kwargs)
    if git_status:
        for change in git_status.split('\n'):
            # change[0] contains staged status
            unstaged_status = change[1]
            path = change[3:]
            # if path contains space it will be surrounded by "
            path = path.strip('"')
            if unstaged_status in ['M', '?']:
                added.append(path)
            elif unstaged_status == 'D':
                deleted.append(path)
            else:
                other.append(path)
    return {'added': added, 'deleted': deleted, 'other': other}


def git_update_gitignore(repo, new_ignore_patterns):
    gitignore_filepath = os.path.join(repo.working_dir, '.gitignore')
    ignored_set = set()
    with open(gitignore_filepath, 'r') as fobj:
        for line in fobj.readlines():
            ignored_set.add(line.strip())
    need_updating = False
    for new_ignore in new_ignore_patterns:
        new_ignore = new_ignore.strip()
        if new_ignore not in ignored_set:
            need_updating = True
            ignored_set.add(new_ignore)
    if need_updating:
        with open(gitignore_filepath, 'w') as fobj:
            new_gitignore_str = '\n'.join(sorted(ignored_set))
            fobj.write(new_gitignore_str)
        repo.git.execute(['git', 'add', gitignore_filepath], **extra_git_kwargs)
        repo.git.execute(['git', 'commit', '-m Update gitignore'], **extra_git_kwargs)


def git_pull_develop_bundle(response, repo, path):
    bundle_path = os.path.join(path, '.develop.bundle')
    with open(bundle_path, 'wb') as fobj:
        for chunk in response.iter_content(512 * 1024):
            if not chunk:
                break
            fobj.write(chunk)
    if not 'develop_bundle' in repo.git.execute(
            ['git', 'remote'], **extra_git_kwargs).split():
        repo.git.execute(
            ['git', 'remote', 'add', 'develop_bundle', bundle_path],
            **extra_git_kwargs)
    repo.git.execute(
        ['git', 'fetch', 'develop_bundle'], **extra_git_kwargs)
    if not 'develop' in repo.git.execute(
            ['git', 'branch'], **extra_git_kwargs).split():
        repo.git.execute(
            ['git', 'checkout', '-bdevelop', 'develop_bundle/develop'],
            **extra_git_kwargs)
    repo.git.execute(['git', 'merge', 'develop_bundle/develop', 'develop'],
                     **extra_git_kwargs)



# -*- coding: utf-8 -*-
import datetime
import os
import platform
import requests
import shutil
import threading
import time
import traceback

from git.exc import GitCommandError

from .utils import (
    ValidationError, is_hidden, is_valid_file_name, resource_path,
    uniform_filepath)
from .sync_helpers import (
    BACKUP_COUNT, LOG_FILENAME, get_site_specific_logger, git_changes,
    git_pull_develop_bundle, extra_git_kwargs)

SYNCABLE_DIRECTORIES = ('templates/', 'static/', 'private/')

IGNORED_FILES = set(['.aldryn', '.aldryn-folder', '.aldryn-sync-lock',
                     '.DS_Store', LOG_FILENAME])
for i in xrange(1, BACKUP_COUNT + 1):
    IGNORED_FILES.add('%s.%d' % (LOG_FILENAME, i))

CHANGES_CHECK_DELAY = 0.5  # 0.5s
SYNC_BACK_EVERY_X_CHECKS = 10  # 5s
MAX_SYNC_RETRY_COUNT = 2


def update_env_path_with_git_bin():
    system = platform.system()
    if system == 'Darwin':
        os.environ['PATH'] = '%s:' % resource_path('resources/mac_osx/bin') + os.environ['PATH']
        os.environ['GIT_EXEC_PATH'] = resource_path('resources/mac_osx/libexec/git-core')
    elif system == 'Windows':
        os.environ['PATH'] = '%s;' % resource_path('resources/windows/bin') + os.environ['PATH']
    else:
        pass  # TODO
update_env_path_with_git_bin()


class GitSyncHandler(object):

    def __init__(self, client, sitename, repo, last_synced_commit,
                 network_error_callback, sync_error_callback,
                 protected_files, protected_file_change_callback,
                 relpath='.', sync_indicator_callback=None):
        self.client = client
        self.sitename = sitename
        self.repo = repo
        self._last_synced_commit = last_synced_commit
        self.relpath = uniform_filepath(relpath)
        self.sync_logger = get_site_specific_logger(sitename, self.relpath)
        self._sync_stopped_event = threading.Event()
        # UI callbacks
        self._network_error_callback = network_error_callback
        self._sync_error_callback = sync_error_callback
        self._sync_indicator_callback = sync_indicator_callback
        # protected files
        self._protected_files = protected_files
        self._overridden_protected_files = set()
        self._protected_file_change_callback = protected_file_change_callback
        # warnings about incorret files
        self._already_notified_incorrect_files = set()
        # sync error dialog handlers
        self._sync_error_dialogs = []

    def start(self):
        self._send_changes_thread = threading.Thread(
            target=self._send_changes_worker, name='Changes sender')
        self._send_changes_thread.start()

    def stop(self):
        try:
            self.client.session.post('/api/v1/sync/%s/sync-log-stop/' % self.sitename)
        except Exception as e:
            # suppress exception, this request isn't that important, sync will
            # eventually be marked as stopped
            print e
        self._sync_stopped_event.set()
        if hasattr(self, '_send_changes_thread'):
            self._send_changes_thread.join()
        self.client._remove_sync_lock(self.relpath)
        if self.sitename in self.client._sync_handlers_cache:
            del self.client._sync_handlers_cache[self.sitename]

    def _send_changes_worker(self):
        iteration_counter = 0
        while not self._sync_stopped_event.isSet():
            start_timestamp = datetime.datetime.now()
            changes = git_changes(self.repo)
            commit = False
            added_files = changes['added']
            for file_rel_path in added_files:
                filepath = os.path.join(self.relpath, file_rel_path)
                file_basename = os.path.basename(filepath)
                if file_basename in IGNORED_FILES:
                    continue
                if (file_rel_path in self._protected_files and
                        file_rel_path not in self._overridden_protected_files):
                    from .client import Client
                    message = Client.PROTECTED_FILE_CHANGE_MESSAGE % filepath
                    self._protected_file_change_callback(message)
                    self._overridden_protected_files.add(file_rel_path)

                syncable = True
                in_sync_dir = file_rel_path.startswith(SYNCABLE_DIRECTORIES)
                if not in_sync_dir:
                    error_msg = ('Not in the syncable directory: %s' %
                                 (', '.join(SYNCABLE_DIRECTORIES)))
                    syncable = False
                elif is_hidden(file_basename):
                    error_msg = 'Hidden files aren\'t synchronized.'
                    syncable = False
                else:
                    def raiser(msg):
                        raise ValidationError(msg)

                    try:
                        syncable = is_valid_file_name(
                            file_basename, logger=raiser)
                    except ValidationError as e:
                        syncable = False
                        error_msg = e.message
                if syncable:
                    commit = True
                    self.repo.git.execute(
                        ['git', 'add', file_rel_path], **extra_git_kwargs)
                elif filepath not in self._already_notified_incorrect_files:
                    self._already_notified_incorrect_files.add(filepath)
                    msg = 'not a syncable file: "%s": %s' % (filepath, error_msg)
                    self._sync_error_callback(msg)
                    self.sync_logger.info(msg)
            deleted_files = changes['deleted']
            for file_rel_path in deleted_files:
                filepath = os.path.join(self.relpath, file_rel_path)
                if not os.path.exists(filepath):
                    try:
                        self.repo.git.execute(
                            ['git', 'rm', file_rel_path], **extra_git_kwargs)
                    except GitCommandError:
                        # Some editors remove the original file and rename
                        # the updated copy. We just run the command during the
                        # process.
                        pass
                    else:
                        commit = True
            if commit:
                for key, values in changes.items():
                    if values:
                        print ('%s:' % key),
                        for value in values:
                            print ('\t%s' % value),
                        print
                self._commit_changes()

            stop_timestamp = datetime.datetime.now()
            time_elapsed_delta = stop_timestamp - start_timestamp
            delay = CHANGES_CHECK_DELAY - time_elapsed_delta.total_seconds()
            if delay > 0:
                time.sleep(delay)
            iteration_counter += 1
            if iteration_counter % SYNC_BACK_EVERY_X_CHECKS == 0:
                self._sync_back_upstream_changes()

    def _commit_changes(self):
        self.sync_logger.debug('Sending changes:\t' + self.repo.git.execute(
            ['git', 'status', '--porcelain'], **extra_git_kwargs))
        try:
            self.repo.git.execute(
                ['git', 'commit', '-m Git Sync'], **extra_git_kwargs)
        except Exception as e:
            print e
            # TODO this sometimes fails, probably due to file being modified again
            # while committing, we just skip this sync turn and try next time
            return

        retry_event = threading.Event()
        exit_loop_event = threading.Event()

        def on_confirm():
            retry_event.set()

        def on_cancel():
            exit_loop_event.set()
            retry_event.set()

        if self._sync_indicator_callback:
            self._sync_indicator_callback()
        try_count = 0
        while not exit_loop_event.isSet():
            try_count += 1
            sync_bundle_path = os.path.join(self.relpath, '.sync.bundle')
            commits_range = '%s..develop' % self._last_synced_commit
            self.repo.git.execute(
                ['git', 'bundle', 'create', sync_bundle_path, commits_range],
                **extra_git_kwargs)
            try:
                with open(sync_bundle_path, 'rb') as fobj:
                    response = self._send_request(files={'content': fobj})
            except (requests.exceptions.ConnectionError,
                    requests.exceptions.Timeout):
                retry_event.clear()
                from .client import Client
                message = Client.SYNC_NETWORK_ERROR_MESSAGE
                self._network_error_callback(
                    message, on_confirm, on_cancel)
                retry_event.wait()
            else:
                if response.ok:
                    # Dismiss previous sync error dialogs
                    for dialog in self._sync_error_dialogs:
                        if dialog and hasattr(dialog, 'dismiss'):
                            dialog.dismiss()
                    self._sync_error_dialogs = []
                    # Updating local "file" remote after successful sync of the commits
                    develop_bundle_path = os.path.join(self.relpath, '.develop.bundle')
                    shutil.move(sync_bundle_path, develop_bundle_path)
                    self.repo.git.execute(
                        ['git', 'fetch', 'develop_bundle'], **extra_git_kwargs)

                    self._last_synced_commit = self.repo.git.execute(
                        ['git', 'rev-parse', 'develop_bundle/develop'], **extra_git_kwargs)
                    exit_loop_event.set()
                elif response.status_code == 409 and try_count <= MAX_SYNC_RETRY_COUNT:
                    # probably upstream has some new commits which we need to pull
                    self.sync_logger.debug(response.content)
                    self._sync_back_upstream_changes()
                else:
                    title = "Sync failed!"
                    if response.status_code in (400, 409):
                        msg = response.content
                    else:
                        base_msg = "Unexpected status code %s" % response.status_code
                        if response.status_code < 500:
                            msg = '\n'.join([base_msg, response.content])
                        else:
                            msg = '\n'.join([base_msg, "Internal Server Error"])
                    sync_error_dialog = self._sync_error_callback(msg, title=title)
                    self._sync_error_dialogs.append(sync_error_dialog)
                    exit_loop_event.set()
            # endwhile
        if self._sync_indicator_callback:
            self._sync_indicator_callback(stop=True)

    def _sync_back_upstream_changes(self):
        try:
            last_synced_commit = self.repo.git.execute(
                ['git', 'rev-parse', 'develop_bundle/develop'],
                **extra_git_kwargs)
            git_sync_params = {'last_synced_commit': last_synced_commit}
            response = self.client.session.get(
                '/api/v1/git-sync/%s/' % self.sitename, params=git_sync_params,
                stream=True, headers={'accept': 'application/octet'})
            if response.status_code == 304:
                pass  # NOT MODIFIED
            elif response.status_code == 200:
                git_pull_develop_bundle(response, self.repo, self.relpath)
        except (requests.exceptions.ConnectionError,
                requests.exceptions.Timeout, GitCommandError):
            stack_trace = traceback.format_exc()
            print 'Suppressing pull update error:\n%s' % stack_trace

    def _send_request(self, *args, **kwargs):
        headers = kwargs.get('headers', {})
        if 'accept' not in headers:
            headers['accept'] = 'text/plain'
        kwargs['headers'] = headers
        response = self.client.session.request(
            'POST', '/api/v1/git-sync/%s/' % self.sitename, *args, **kwargs)
        return response



# -*- coding: utf-8 -*-
from aldryn_client.serialize import Dumper
from django.conf import settings
from django.core.management.base import BaseCommand
from django.utils.translation import get_language
import os


class Command(BaseCommand):
    def handle(self, outfile=None, language=None, **options):
         if outfile is None:
             print "You must specify a file to write to as first argument"
             return
         if language is None:
             print "You must specify a language code (eg: en) as second argument"
             return
         datadir = os.path.join(os.path.dirname(outfile), 'data')
         dumper = Dumper(datadir, language, getattr(settings, 'ALDRYN_DUMPDATA_FOLLOW', []))
         dumper.dump(outfile)





# -*- coding: utf-8 -*-
from collections import defaultdict
from distutils.version import StrictVersion
import hashlib
import shutil
import glob

import os
import yaml


# YAML STUFF
class Tracker(object):
    def __init__(self):
        self.contexts = []

    def push(self, thing):
        for context in self.contexts:
            context[thing.__class__].append(thing)

    def __enter__(self):
        context = defaultdict(list)
        self.contexts.append(context)
        return context

    def __exit__(self, exc_type, exc_val, exc_tb):
        del self.contexts[-1]


class Trackable(object):
    tracker = Tracker()

    def __init__(self):
        self.tracker.push(self)


class Include(Trackable):
    def __init__(self, path):
        super(Include, self).__init__()
        self.path = path


class ListInclude(Include): pass


class File(Include): pass


class LiteralInclude(Include): pass


class Model(Trackable):
    def __init__(self, app_label, model_name):
        super(Model, self).__init__()
        self.app_label = app_label
        self.model_name = model_name

    def load(self):
        from django.db.models.loading import get_model

        return get_model(self.app_label, self.model_name)


def include_constructor(loader, node):
    value = loader.construct_scalar(node).lstrip('/')
    with open(value) as fobj:
        return yaml.safe_load(fobj)


def include_representer(dumper, data):
    return dumper.represent_scalar(u'!include', u'%s' % data.path)


def list_include_constructor(loader, node):
    value = loader.construct_scalar(node).lstrip('/')
    data = []
    for fname in glob.glob(value):
        with open(fname) as fobj:
            data += yaml.safe_load(fobj)
    return data


def list_include_representer(dumper, data):
    return dumper.represent_scalar(u'!list-include', data.path)


def file_constructor(loader, node):
    path = loader.construct_scalar(node)
    return File(path)


def file_representer(dumper, data):
    return dumper.represent_scalar(u'!file', data.path)


def literal_include_constructor(loader, node):
    with open(loader.construct_scalar(node).lstrip('/')) as fobj:
        return fobj.read()


def literal_include_representer(dumper, data):
    return dumper.represent_scalar(u'!literal-include', data.path)


def model_representer(dumper, data):
    return dumper.represent_scalar(u'!model', '%s.%s' % (data.app_label, data.model_name))


def model_constructor(loader, node):
    name_string = loader.construct_scalar(node)
    app_label, model_name = name_string.split('.')
    return Model(app_label, model_name)


__registerered = False


def register_yaml_extensions():
    global __registerered
    if __registerered:
        return

    yaml.add_constructor(u'!include', include_constructor, yaml.SafeLoader)
    yaml.add_constructor(u'!include-list', list_include_constructor, yaml.SafeLoader)
    yaml.add_constructor(u'!literal-include', literal_include_constructor, yaml.SafeLoader)
    yaml.add_constructor(u'!file', file_constructor, yaml.SafeLoader)
    yaml.add_constructor(u'!model', model_constructor, yaml.SafeLoader)
    yaml.add_representer(Include, include_representer, yaml.SafeDumper)
    yaml.add_representer(ListInclude, list_include_representer, yaml.SafeDumper)
    yaml.add_representer(File, file_representer, yaml.SafeDumper)
    yaml.add_representer(LiteralInclude, literal_include_representer, yaml.SafeDumper)
    yaml.add_representer(Model, model_representer, yaml.SafeDumper)

    __registerered = True

# DUMP


class Dumper(object):
    def __init__(self, datadir, language, follow=None):
        from django.utils.translation import activate

        activate(language)
        register_yaml_extensions()
        self.datadir = datadir
        if os.path.exists(self.datadir):
            shutil.rmtree(self.datadir)
        os.mkdir(self.datadir)
        self.file_count = 0
        self.language = language
        self.follow = defaultdict(list)
        if follow:
            for key, value in (thing.split('.', 1) for thing in follow):
                self.follow[key].append(value)
        from cms import models

        self.queryset = models.Page.objects.root().public()
        self.file_cache = {}

    def dump(self, filename):
        data = self.get_pages()
        with open(filename, 'w') as fobj:
            yaml.safe_dump(data, fobj)

    def get_pages(self):
        data = []
        for page in self.queryset:
            data.append(self.serialize_page(page))
        return data

    def serialize_page(self, page):
        return {
            'path': page.get_absolute_url(),
            'name': page.get_title(),
            'template': page.template,
            'placeholders': [self.serialize_placeholder(page, ph) for ph in page.placeholders.all()],
            'children': [self.serialize_page(child) for child in page.get_children()],
        }

    def serialize_placeholder(self, page, placeholder):
        plugins = self.dump_plugins(page, placeholder)
        return {
            'name': placeholder.slot,
            'plugins': Include(plugins)
        }

    def dump_plugins(self, page, placeholder):
        filename = os.path.join(self.datadir, '%s_%s.yaml' % (placeholder.slot, page.pk))
        data = [self.serialize_plugin(plugin) for plugin in
                placeholder.cmsplugin_set.filter(language=self.language, parent__isnull=True).order_by('position')]
        with open(filename, 'w') as fobj:
            yaml.safe_dump(data, fobj)
        return filename

    def serialize_plugin(self, plugin):
        from django.forms.models import model_to_dict

        instance = plugin.get_plugin_instance()[0]
        raw_data = model_to_dict(instance)
        raw_data.pop('cmsplugin_ptr', None)
        del raw_data['id']
        raw_data['plugin_type'] = plugin.plugin_type
        raw_data['-relations'] = []
        raw_data['-children'] = []
        self.post_process_files(raw_data)
        self.post_process_relations(instance, plugin, raw_data)
        self.post_process_children(instance, plugin, raw_data)
        return raw_data

    def post_process_files(self, raw_data):
        from django.db.models.fields.files import FieldFile, ImageFieldFile

        for key, value in raw_data.items():
            if isinstance(value, (FieldFile, ImageFieldFile)):
                if value:
                    data = ''
                    for chunk in value.chunks():
                        data += chunk
                        #de-dup
                    checksum = hashlib.md5(data).hexdigest()
                    if checksum not in self.file_cache:
                        filename = os.path.join(self.datadir, '_file_%s' % self.file_count)
                        self.file_count += 1
                        with open(filename, 'wb') as target:
                            target.write(data)
                        self.file_cache[checksum] = filename
                    raw_data[key] = File(self.file_cache[checksum])
                else:
                    raw_data[key] = None

    def post_process_relations(self, instance, plugin, raw_data):
        for field in self.follow['%s:%s' % (instance._meta.app_label, instance._meta.module_name)]:
            objects = getattr(instance, field).all()
            raw_data['-relations'] += [self.serialize_plugin_relation(instance, obj) for obj in objects]

    def serialize_plugin_relation(self, plugin, obj):
        from django.forms.models import model_to_dict
        from django.db.models.fields.related import ForeignKey

        data = model_to_dict(obj, fields=obj._meta.get_all_field_names())
        for field_name in obj._meta.get_all_field_names():
            field = obj._meta.get_field_by_name(field_name)[0]
            if isinstance(field, ForeignKey):
                if getattr(obj, field_name) == plugin:
                    del data[field_name]
                    data['-field'] = field_name
            elif field_name not in data:
                data[field_name] = field.value_from_object(obj)
        del data['id']
        self.post_process_files(data)
        data['-model'] = Model(obj._meta.app_label, obj.__class__.__name__)
        return data

    def post_process_children(self, instance, plugin, raw_data):
        pass

# LOAD

class Loader(object):
    def __init__(self, language):
        from django.utils.translation import activate

        activate(language)
        register_yaml_extensions()

    def syncdb(self):
        from django.core.management import call_command
        from django.conf import settings

        call_command('syncdb', interactive=False)
        if 'south' in settings.INSTALLED_APPS:
            call_command('migrate', interactive=False)

    def load(self, filename):
        self.syncdb()
        from cms.models import Page, Placeholder
        self.placeholder_does_not_exist = Placeholder.DoesNotExist

        if Page.objects.exists():
            print "Non-empty database, aborting"
            return
        with open(filename) as fobj:
            data = yaml.safe_load(fobj)
            for page in data:
                self.load_page(page)

    def load_page(self, data, parent=None):
        from cms.api import create_page
        from cms.models import Page
        from django.utils.translation import get_language

        if parent:
            parent = Page.objects.get(pk=parent.pk)
        page = create_page(data['name'], data['template'], get_language(), parent=parent,
                           in_navigation=True, published=True)
        for placeholder in data['placeholders']:
            self.load_placeholder(placeholder, page)
        for child in data['children']:
            self.load_page(child, page)
        page.publish()

    def load_placeholder(self, data, page):
        try:
            placeholder = page.placeholders.get(slot=data['name'])
        except self.placeholder_does_not_exist:
            placeholder = page.placeholders.create(slot=data['name'])
        for plugin in data['plugins']:
            self.load_plugin(plugin, placeholder)

    def load_plugin(self, data, placeholder, parent=None):
        from cms.api import add_plugin
        from django.utils.translation import get_language

        plugin_type = data.pop('plugin_type')
        relations = data.pop('-relations')
        children = data.pop('-children')
        self.pre_process_files(data)
        plugin = add_plugin(placeholder, plugin_type, get_language(), target=parent, **data)
        for child in children:
            self.load_plugin(child, placeholder, plugin)
        for relation in relations:
            self.load_relation(relation, plugin)

    def pre_process_files(self, data):
        from django.core.files.base import File as DjangoFile

        for key, value in data.items():
            if isinstance(value, File):
                data[key] = DjangoFile(open(value.path))

    def load_relation(self, data, plugin):
        model = data.pop('-model').load()
        plugin_field = data.pop('-field')
        data[plugin_field] = plugin
        model.objects.create(**data)



# -*- coding: utf-8 -*-
import datetime
import os
import requests
import threading
import time

from watchdog.events import (
    DirCreatedEvent, DirDeletedEvent, FileCreatedEvent, FileDeletedEvent,
    FileSystemEventHandler)

from .utils import uniform_filepath
from .sync_helpers import (
    EventsBuffer, FileHashesCache, ProceededEventsQueue, SyncEvent,
    get_site_specific_logger, BACKUP_COUNT, LOG_FILENAME)

# Amount of time during which we collect events and perform heuristics
# to reduce the number of requests
TIME_DELTA_IN_SECONDS = 0.5

# Waiting twice (and a bit) as long as it takes to consider subsequent events
# as a single action e.g. directory move, file save (by 'moving/renaming'),
# to collect all actions from a single "batch" into (preferably) single request
COLLECT_TIME_DELTA_IN_SECONDS = TIME_DELTA_IN_SECONDS * 2.01

IGNORED_FILES = set(['.aldryn', '.aldryn-folder', '.aldryn-sync-lock',
                     '.DS_Store', LOG_FILENAME])
for i in xrange(1, BACKUP_COUNT + 1):
    IGNORED_FILES.add('%s.%d' % (LOG_FILENAME, i))

###############################################################################
# End of constants
###############################################################################


class SyncEventHandler(FileSystemEventHandler):

    def __init__(self, session, sitename, observer_stopped,
                 network_error_callback, sync_error_callback,
                 protected_files, protected_file_change_callback,
                 relpath='.', sync_indicator_callback=None):
        self.session = session
        self.sitename = sitename
        self.relpath = uniform_filepath(relpath)
        self.sync_logger = get_site_specific_logger(sitename, self.relpath)
        self._observer_stopped = observer_stopped
        self._network_error_callback = network_error_callback
        self._sync_error_callback = sync_error_callback
        self._sync_indicator_callback = sync_indicator_callback
        self._protected_files = protected_files
        self._overridden_protected_filed = set()
        self._protected_file_change_callback = protected_file_change_callback

        self._recently_modified_file_hashes = {}

        self._events_buffer = EventsBuffer(self.relpath)

        self._proceeded_events_queue = ProceededEventsQueue()

        self._file_hashes_cache = FileHashesCache(self.relpath)
        self._file_hashes_cache.update_hashes()

        # Due to the asynchronous nature of the file system's events
        # there should be only one 'sending requests worker'.
        # (e.g. lagging 'create request' followed by a 'delete' one
        # will fail because the fail wasn't created yet)
        self._send_requests_thread = threading.Thread(
            target=self._send_requests_worker, name='Requests sender')
        self._send_requests_thread.start()

        self._collect_events_thread = threading.Thread(
            target=self._collect_events_worker, name='Events collector')
        self._collect_events_thread.start()

    def _collect_events_worker(self):
        while not self._observer_stopped.isSet():
            start_timestamp = datetime.datetime.now()

            oldest_events_buffer = (
                self._events_buffer.filter_out_oldest_events_buffer(
                    TIME_DELTA_IN_SECONDS))
            oldest_events_buffer.log_state(self.sync_logger)
            reduced_sync_events = oldest_events_buffer.get_reduced_events_list(
                self._file_hashes_cache.is_file_changed,
                self.sync_logger)
            for sync_event in reduced_sync_events:
                self.sync_logger.debug(
                    'Queueing the event:\t' + repr(sync_event))
                self._proceeded_events_queue.put_event(sync_event)

            stop_timestamp = datetime.datetime.now()
            time_elapsed_delta = stop_timestamp - start_timestamp
            delay = COLLECT_TIME_DELTA_IN_SECONDS - time_elapsed_delta.total_seconds()
            if delay > 0:
                time.sleep(delay)

    def _send_requests_worker(self):
        while not self._observer_stopped.isSet():
            sync_event = self._proceeded_events_queue.get_event(
                timeout=TIME_DELTA_IN_SECONDS)
            if sync_event:
                from .client import Client
                rel_src_path = sync_event.rel_src_path
                if (rel_src_path in self._protected_files and
                        rel_src_path not in self._overridden_protected_filed):
                    message = Client.PROTECTED_FILE_CHANGE_MESSAGE % sync_event.src_path
                    self._protected_file_change_callback(message)
                    self._overridden_protected_filed.add(rel_src_path)

                self.sync_logger.debug(
                    'Sending request for event:\t' + repr(sync_event))
                msg, method, kwargs, fobj = sync_event.prepare_request()
                print msg

                retry_event = threading.Event()
                exit_loop_event = threading.Event()

                def on_confirm():
                    retry_event.set()

                def on_cancel():
                    exit_loop_event.set()
                    retry_event.set()

                if self._sync_indicator_callback:
                    self._sync_indicator_callback()
                while not exit_loop_event.isSet():
                    if fobj:  # reseting the read state of the sending file
                        fobj.seek(0)
                    try:
                        self._send_request(method, **kwargs)
                    except (requests.exceptions.ConnectionError,
                            requests.exceptions.Timeout):
                        retry_event.clear()
                        message = Client.SYNC_NETWORK_ERROR_MESSAGE % sync_event.src_path
                        self._network_error_callback(
                            message, on_confirm, on_cancel)
                        retry_event.wait()
                    else:
                        if fobj:
                            fobj.close()
                        exit_loop_event.set()
                if self._sync_indicator_callback:
                    self._sync_indicator_callback(stop=True)

    def _send_request(self, method, *args, **kwargs):
        headers = kwargs.get('headers', {})
        if 'accept' not in headers:
            headers['accept'] = 'text/plain'
        kwargs['headers'] = headers
        response = self.session.request(
            method, '/api/v1/sync/%s/' % self.sitename, *args, **kwargs)
        if not response.ok:
            title = "Sync failed!"
            if response.status_code == 400:
                msg = response.content
            else:
                base_msg = "Unexpected status code %s" % response.status_code
                if response.status_code < 500:
                    msg = '\n'.join([base_msg, response.content])
                else:
                    msg = '\n'.join([base_msg, "Internal Server Error"])
            self._sync_error_callback(msg, title=title)

    def dispatch(self, raw_event):
        now = datetime.datetime.now()
        sync_event = SyncEvent(raw_event, now, self.relpath)
        if sync_event.base_src_path not in IGNORED_FILES:
            super(SyncEventHandler, self).dispatch(sync_event)

    def on_moved(self, sync_event):
        if sync_event.is_src_path_syncable:
            if sync_event.is_dest_path_syncable:
                self._events_buffer.set_moved_event(sync_event)
            else:
                self.sync_logger.debug(
                    'Moved outside of the syncable area, removing: %r' % sync_event)
                if sync_event.is_directory:
                    raw_delete_event = DirDeletedEvent(sync_event.src_path)
                else:
                    raw_delete_event = FileDeletedEvent(sync_event.src_path)
                delete_event = SyncEvent(
                    raw_delete_event, sync_event.timestamp, sync_event.relpath)
                self.on_deleted(delete_event)
        elif sync_event.is_dest_path_syncable:
            self.sync_logger.debug(
                'Moved into the syncable area, creating: %r' % sync_event)
            if sync_event.is_directory:
                raw_create_event = DirCreatedEvent(sync_event.dest_path)
            else:
                raw_create_event = FileCreatedEvent(sync_event.dest_path)
            create_event = SyncEvent(
                raw_create_event, sync_event.timestamp, sync_event.relpath)
            self.on_created(create_event)

    def _log_not_a_syncable_file(self, sync_event):
        self.sync_logger.debug(
            'not a syncable file: "%s": %s' %
            (sync_event, sync_event.not_syncable_src_path_reason))
        print ('Cannot sync file "%s": %s' %
               (sync_event.src_path,
                sync_event.not_syncable_src_path_reason))

    def on_created(self, sync_event):
        if not sync_event.is_src_path_syncable:
            self._log_not_a_syncable_file(sync_event)
            return
        if sync_event.is_directory:
            # check if the directory has a content, if so create the stuff
            try:
                for thing in os.listdir(sync_event.src_path):
                    raw_create_event = FileCreatedEvent(
                        os.path.join(sync_event.src_path, thing))
                    create_event = SyncEvent(
                        raw_create_event, sync_event.timestamp, sync_event.relpath)
                    if create_event.base_src_path not in IGNORED_FILES:
                        self.on_created(create_event)
            except os.OSError as e:
                self.sync_logger.error(
                    'Accessing created directory "%s" raised: %r.' %
                    (sync_event.src_path, e))
        else:
            self._events_buffer.set_created_event(sync_event)

    def on_deleted(self, sync_event):
        if not sync_event.is_src_path_syncable:
            self._log_not_a_syncable_file(sync_event)
            return
        self._events_buffer.set_deleted_event(sync_event)

    def on_modified(self, sync_event):
        if sync_event.is_directory:
            # This doesn't mean that the directory was renamed
            # it just means that the content of it was modified
            # which we handle in their own events
            return
        if not sync_event.is_src_path_syncable:
            self._log_not_a_syncable_file(sync_event)
            return
        else:
            self._events_buffer.set_modified_event(sync_event)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from setuptools import setup, find_packages
from aldryn_client import __version__


INSTALL_REQUIRES = [
    'GitPython==0.3.2.RC1',
    'requests>=1',
    'docopt',
    'pyyaml',
]
DEPENDENCY_LINKS = [
    'https://bitbucket.org/pygame/pygame/get/df48571.zip#egg=pygame-1.9.2',
]
EXTRAS_REQUIRE = {
    'gui': [
        'Kivy==1.7.2',
        'plyer==1.1.2',
        'pygame',
    ],
}
try:
    import json
except ImportError:
    INSTALL_REQUIRES.append('simplejson')

CLASSIFIERS = [
    'Development Status :: 5 - Production/Stable',
    'Environment :: Console',
    'Framework :: Django',
    'Intended Audience :: Developers',
    'License :: OSI Approved :: BSD License',
    'Operating System :: OS Independent',
    'Programming Language :: Python',
    'Programming Language :: Python :: 2.6',
    'Programming Language :: Python :: 2.7',
    'Topic :: Software Development',
]

setup(
    name='aldryn-client',
    version=__version__,
    description='The command-line client for the Aldryn Cloud',
    author='Divio AG',
    author_email='aldryn@divio.ch',
    url='http://www.aldryn.com/',
    packages=find_packages(),
    include_package_data=True,
    zip_safe=False,
    license='BSD',
    platforms=['OS Independent'],
    install_requires=INSTALL_REQUIRES,
    extras_require=EXTRAS_REQUIRE,
    dependency_links=DEPENDENCY_LINKS,
    entry_points="""
    [console_scripts]
    aldryn = aldryn_client.cli:main
    """,
)



# -*- coding: utf-8 -*-
from cStringIO import StringIO
import hashlib
import imp
import json
import os
import re
import shutil
import subprocess
import sys
import tarfile
import tempfile
import time
import yaml

from .serialize import register_yaml_extensions, Trackable, File

FILENAME_BASIC_RE = re.compile(r'^[a-zA-Z0-9_]+[a-zA-Z0-9._-]*(@2x)?\.[a-zA-Z]{2,4}$')
ALLOWED_EXTENSIONS = [
    '.css',
    '.gif',
    '.htc',
    '.htm',
    '.html',
    '.ico',
    '.jpeg',
    '.jpg',
    '.js',
    '.json',
    '.less',
    '.map',
    '.png',
    '.rb',
    '.sass',
    '.scss',
    '.svg',
    '.webm',
    # font formats
    '.eot',
    '.ttf',
    '.woff',
    # text formats
    '.txt',
    '.md',
    '.rst',
    # document formats
    '.pdf',
    '.ps',
    '.djv',
    '.djvu',
    '.doc',
    '.docx',
    '.ppt',
    '.pptx',
    '.xls',
    '.xlsx',
]

BOILERPLATE_REQUIRED = [
    'name',
    ('author', [
        'name',
    ]),
    'version',
    'description',
    ('license', [
        'name',
    ]),
    'templates',
]

BOILERPLATE_REQUIRED_FILEPATHS = [
    os.path.join('templates', 'base.html'),
]

APP_REQUIRED = [
    'name',
    ('author', [
        'name',
    ]),
    'package-name',
    'description',
    ('license', [
        'name',
    ]),
    'installed-apps',
]

VALID_LICENSE_FILENAMES = [
    'LICENSE.txt',
    'LICENSE',
    'license.txt',
    'license',
]


class ValidationError(Exception):
    pass


def _validate(config, required, path):
    license_exists = False
    for valid_license_filename in VALID_LICENSE_FILENAMES:
        license_exists |= os.path.exists(
            os.path.join(path, valid_license_filename))
    if not license_exists:
        return (False, "Required LICENSE.txt file not found")
    valid = (True, "Configuration file is valid")
    for thing in required:
        if isinstance(thing, tuple):
            key, values = thing
        else:
            key, values = thing, []

        if key not in config:
            valid = (False, "Required key %r not found in config" % key)

        for subkey in values:
            if subkey not in config[key]:
                valid = (False, "Required sub key %r in %r not found in config" % (subkey, key))
    return valid


def validate_app_config(config, path):
    if 'version' in config:
        print ("WARNING! Use of the 'version' field in the 'addon.json' is deprecated, "
               "please remove it.")
    aldryn_config_path = os.path.abspath(os.path.join(path, 'aldryn_config.py'))
    if os.path.exists(aldryn_config_path):
        tempdir = tempfile.mkdtemp(prefix='tmp_aldryn_client_')
        try:
            shutil.copy(aldryn_config_path, tempdir)
            filepath = os.path.join(tempdir, 'aldryn_config.py')
            try:
                orig_err = sys.stderr
                sys.stderr = StringIO()
                # suppressing "RuntimeWarning: Parent module 'aldryn_config' not found while handling absolute import"
                module = imp.load_source('aldryn_config.config_%s' % int(time.time()), filepath)
                sys.stderr = orig_err
                # checking basic functionality of the Form
                form = module.Form({})
                form.is_valid()
            except Exception:
                import traceback
                error_msg = traceback.format_exc()
                return (False, "Exception in aldryn_config.py:\n\n%s" % error_msg)
        finally:
            shutil.rmtree(tempdir)
    return _validate(config, APP_REQUIRED, path)


def validate_boilerplate_config(config, path):
    for required_filepath in BOILERPLATE_REQUIRED_FILEPATHS:
        dirpath = os.path.join(path, required_filepath)
        if not os.path.exists(dirpath):
            msg = 'Required file "%s" not found' % required_filepath
            return (False, msg)
    (valid, msg) = _validate(config, BOILERPLATE_REQUIRED, path)
    if not valid:
        return (False, msg)
    # check templates
    data = config.get('templates', [])
    template_valid = True
    if not isinstance(data, list):
        template_valid = False
    else:
        for template in data:
            if not isinstance(template, list):
                template_valid = False
            elif len(template) != 2:
                template_valid = False
    if not template_valid:
        msg = "Templates must be a list of lists of two items"
        return (False, msg)

    # check protected
    protected = config.get('protected', [])
    valid = True
    if not isinstance(protected, list):
        valid = False
        msg = "Protected files must be a list"
    else:
        errors = []
        for filename in protected:
            filepath = os.path.join(path, filename)
            if filepath.endswith('*'):
                continue
            if not os.path.exists(filepath):
                valid = False
                errors.append("Protected file %r not found" % filename)
        if errors:
            msg = os.linesep.join(errors)
    return (valid, msg)


def tar_add_stringio(tar, string_io, name):
    info = tarfile.TarInfo(name=name)
    string_io.seek(0, os.SEEK_END)
    info.size = string_io.tell()
    string_io.seek(0)
    tar.addfile(tarinfo=info, fileobj=string_io)


def is_valid_file_name(name, logger=None):
    if not FILENAME_BASIC_RE.match(name):
        if logger:
            logger("File name %r is not valid, ignoring." % name)
        return False
    ext = os.path.splitext(name)[-1]
    if ext not in ALLOWED_EXTENSIONS:
        if logger:
            logger("File extension %r is not allowed, ignoring." % ext)
        return False
    return True


def filter_static_files(tarinfo):
    if not tarinfo.isfile():
        return tarinfo
    basename = os.path.basename(tarinfo.name)
    if is_valid_file_name(basename):
        return tarinfo
    else:
        return None


def filter_sass_files(tarinfo):
    basename = os.path.basename(tarinfo.name)
    if tarinfo.isfile():
        if is_valid_file_name(basename):
            return tarinfo
        else:
            return None
    elif basename.startswith('.'):
        return None
    else:
        return tarinfo


def filter_template_files(tarinfo):
    if not tarinfo.isfile():
        return tarinfo
    basename = os.path.basename(tarinfo.name)
    ext = os.path.splitext(basename)[1]
    if ext == '.html':
        return tarinfo
    else:
        return None


def _get_license_filename(path):
    for valid_license_filename in VALID_LICENSE_FILENAMES:
        filepath = os.path.join(path, valid_license_filename)
        if os.path.exists(filepath):
            return filepath


def load_license(path):
    filepath = _get_license_filename(path)
    if filepath:
        with open(filepath, 'r') as f:
            license = f.read()
            return license


def bundle_boilerplate(config, data, path, extra_file_paths, **complex_extra):
    register_yaml_extensions()
    fileobj = StringIO()
    tar = tarfile.open(mode='w:gz', fileobj=fileobj)
    config_fileobj = StringIO()
    json.dump(config, config_fileobj)
    tar_add_stringio(tar, config_fileobj, 'boilerplate.json')
    data_fileobj = StringIO()
    yaml.safe_dump(data, data_fileobj)
    tar_add_stringio(tar, data_fileobj, 'data.yaml')
    license_filepath = _get_license_filename(path)
    if license_filepath:
        tar.add(license_filepath, 'LICENSE.txt')
    for extra_path in extra_file_paths:
        tar.add(extra_path)
    for key, value in complex_extra.items():
        dirpath = os.path.join(path, key)
        if os.path.exists(dirpath):
            tar.add(key, filter=value)
    tar.close()
    fileobj.seek(0)
    return fileobj


def get_package_version(path):
    devnull = open(os.devnull, 'w')
    try:
        version = subprocess.check_output(
            ['python', 'setup.py', '--version'], cwd=path, stderr=devnull)
    finally:
        devnull.close()
    return version.strip()


def bundle_package(workspace, tar, path):
    devnull = open(os.devnull, 'w')
    try:
        subprocess.check_call(['python', 'setup.py', 'sdist', '-d', workspace],
                              cwd=path, stdout=devnull, stderr=devnull)
    finally:
        devnull.close()
    egg_file = os.path.join(workspace, os.listdir(workspace)[0])
    tar.add(egg_file, arcname='package.tar.gz')


def bundle_app(config, script, path):
    register_yaml_extensions()
    fileobj = StringIO()
    tar = tarfile.open(mode='w:gz', fileobj=fileobj)
    config_fileobj = StringIO()
    json.dump(config, config_fileobj)
    tar_add_stringio(tar, config_fileobj, 'addon.json')
    script_fileobj = StringIO(script)
    license_filepath = _get_license_filename(path)
    if license_filepath:
        tar.add(license_filepath, 'LICENSE.txt')
    if os.path.exists('aldryn_config.py'):
        tar_add_stringio(tar, script_fileobj, 'aldryn_config.py')
        # add actual package
    distdir = tempfile.mkdtemp(prefix='tmp_aldryn_client_')
    try:
        bundle_package(distdir, tar, path)
    finally:
        shutil.rmtree(distdir)
    version = get_package_version(path)
    version_fobj = StringIO(version)
    info = tarfile.TarInfo(name='VERSION')
    info.size = len(version_fobj.getvalue())
    tar.addfile(info, fileobj=version_fobj)
    tar.close()
    fileobj.seek(0)
    return fileobj


def hashfile(fd, blocksize=65536):
    hasher = hashlib.sha256()
    buf = fd.read(blocksize)
    while len(buf) > 0:
        hasher.update(buf)
        buf = fd.read(blocksize)
    return hasher.digest()


def uniform_filepath(filepath):
    filepath = os.path.abspath(filepath)
    filepath = os.path.realpath(filepath)
    filepath = filepath.rstrip(os.sep)
    return filepath


def is_hidden(path):
    filename = os.path.basename(path)
    return filename.startswith('.')


def is_inside_dir(path, parent_dir):
    path = os.path.join(parent_dir, path)
    path = uniform_filepath(path)
    return path.startswith(parent_dir)


def filter_bad_paths(members, parent_dir):
    parent_dir = uniform_filepath(parent_dir)
    for finfo in members:
        if (not finfo.issym() and not finfo.islnk() and
                is_inside_dir(finfo.path, parent_dir)):
            yield finfo


def resource_path(relative_path):
    """
    Get absolute path of the resource, works for dev and for PyInstaller
    """
    try:
        # PyInstaller creates a temp folder and stores path in _MEIPASS
        base_path = os.sys._MEIPASS
    except Exception:
        base_path = os.path.dirname(__file__)

    return os.path.join(base_path, relative_path)


def cli_confirm(question, message=None, default=None):
    if message:
        print message

    if default is None:
        default_answer_str = '[y/n]'
    else:
        default = bool(default)
        if default is True:
            default_answer_str = '[Y/n]'
        elif default is False:
            default_answer_str = '[y/N]'

    question = '%s %s' % (question.rstrip(), default_answer_str)
    while True:
        answer = raw_input(question)
        if answer.lower() == 'y':
            return True
        elif answer.lower() == 'n':
            return False
        elif default is not None and answer == '':
            return default
        else:
            print "Invalid answer, please type either y or n"


def load_boilerplate_config(path):
    from .client import Client
    boilerplate_filename_json = os.path.join(path, Client.BOILERPLATE_FILENAME_JSON)
    boilerplate_filename_yaml = os.path.join(path, Client.BOILERPLATE_FILENAME_YAML)
    boilerplate_filename = None
    load_json_config = False
    if os.path.exists(boilerplate_filename_yaml):
        boilerplate_filename = boilerplate_filename_yaml
    if os.path.exists(boilerplate_filename_json):
        if boilerplate_filename is None:
            boilerplate_filename = boilerplate_filename_json
            load_json_config = True
        else:
            msg = "Please provide only one config file ('%s' or '%s')" % (
                Client.BOILERPLATE_FILENAME_JSON,
                Client.BOILERPLATE_FILENAME_YAML)
            return (False, msg)
    if boilerplate_filename is None:
        msg = "Neither file '%s' nor '%s' were found." % (
            Client.BOILERPLATE_FILENAME_JSON, Client.BOILERPLATE_FILENAME_YAML)
        return (False, msg)
    extra_file_paths = []
    with open(boilerplate_filename) as fobj:
        try:
            if load_json_config:
                config = json.load(fobj)
            else:
                with Trackable.tracker as extra_objects:
                    config = yaml.safe_load(fobj)
                    extra_file_paths.extend([f.path for f in extra_objects[File]])
        except (yaml.YAMLError, ValueError) as e:
            return (False, repr(e))
        return (True, (config, extra_file_paths))


def load_app_config(path):
    from .client import Client
    app_filename_json = os.path.join(path, Client.APP_FILENAME_JSON)
    app_filename_yaml = os.path.join(path, Client.APP_FILENAME_YAML)
    addon_filename_json = os.path.join(path, Client.ADDON_FILENAME_JSON)
    addon_filename_yaml = os.path.join(path, Client.ADDON_FILENAME_YAML)
    app_filenames = []
    load_json_config = False
    if os.path.exists(app_filename_yaml):
        app_filenames.append(app_filename_yaml)
    if os.path.exists(addon_filename_yaml):
        app_filenames.append(addon_filename_yaml)
    if os.path.exists(app_filename_json):
        app_filenames.append(app_filename_json)
        load_json_config = True
    if os.path.exists(addon_filename_json):
        app_filenames.append(addon_filename_json)
        load_json_config = True
    if len(app_filenames) == 0:
        msg = "File '%s' not found." % Client.ADDON_FILENAME_JSON
        return (False, msg)
    elif len(app_filenames) == 1:
        app_filename = app_filenames[0]
    else:
        msg = "Please provide only one config file (%s)" % (
            ' or '.join(map(os.path.basename, app_filenames)),)
        return (False, msg)
    with open(app_filename) as fobj:
        try:
            if load_json_config:
                config = json.load(fobj)
            else:
                config = yaml.safe_load(fobj)
        except (yaml.YAMLError, ValueError) as e:
            return (False, repr(e))
    return (True, config)
